package main

import (
	"encoding/json"
	"fmt"
	"reflect"
	"strconv"
)

type Meta struct {
	data      any
	data_type string
}

func (m *Meta) Set(value any) bool {
	m.data = value
	m.data_type = reflect.TypeOf(value).Name()
	switch m.data_type {
	case "string", "int32", "int64", "uint32", "uint64", "float32", "float64", "bool":
		return true
	case "[]string", "[]int32", "[]int64", "[]uint32", "[]uint64", "[]float32", "[]float64", "[]bool":
		return true
	case "map[string]string", "map[string]int32", "map[string]int64", "map[string]uint32", "map[string]uint64", "map[string]float32", "map[string]float64", "map[string]bool":
		return true
	case "[]Meta", "map[string]Meta":
		return true
	case "[]*Meta":
		v := []Meta{}
		for _, om := range value.([]*Meta) {
			v = append(v, *om)
		}
		m.data = v
		m.data_type = "[]Meta"
		return true
	case "map[string]*Meta":
		v := map[string]Meta{}
		for key, om := range value.(map[string]*Meta) {
			v[key] = *om
		}
		m.data = v
		m.data_type = "map[string]Meta"
		return true
	default:
		m.data = nil
		m.data_type = ""
		return false
	}
}

func (m *Meta) Complex() bool {
	switch m.data_type {
	case "[]string", "[]int32", "[]int64", "[]uint32", "[]uint64", "[]float32", "[]float64", "[]bool":
		return true
	case "map[string]string", "map[string]int32", "map[string]int64", "map[string]uint32", "map[string]uint64", "map[string]float32", "map[string]float64", "map[string]bool":
		return true
	case "[]Meta", "map[string]Meta":
		return true
	default:
		return false
	}
}

func (m *Meta) IsVec() bool {
	switch m.data_type {
	case "[]string", "[]int32", "[]int64", "[]uint32", "[]uint64", "[]float32", "[]float64", "[]bool":
		return true
	case "[]Meta":
		return true
	default:
		return false
	}
}

func (m *Meta) IsMap() bool {
	switch m.data_type {
	case "map[string]string", "map[string]int32", "map[string]int64", "map[string]uint32", "map[string]uint64", "map[string]float32", "map[string]float64", "map[string]bool":
		return true
	case "map[string]Meta":
		return true
	default:
		return false
	}
}

func (m *Meta) Nested() bool {
	switch m.data_type {
	case "[]Meta", "map[string]Meta":
		return true
	default:
		return false
	}
}

func (m *Meta) Len() int {
	switch m.data_type {
	case "[]string":
		v := m.data.([]string)
		return len(v)
	case "[]int32":
		v := m.data.([]int32)
		return len(v)
	case "[]int64":
		v := m.data.([]int64)
		return len(v)
	case "[]uint32":
		v := m.data.([]uint32)
		return len(v)
	case "[]uint64":
		v := m.data.([]uint64)
		return len(v)
	case "[]float32":
		v := m.data.([]float32)
		return len(v)
	case "[]float64":
		v := m.data.([]float64)
		return len(v)
	case "[]bool":
		v := m.data.([]bool)
		return len(v)
	case "[]Meta":
		v := m.data.([]Meta)
		return len(v)
	case "map[string]string":
		v := m.data.(map[string]string)
		return len(v)
	case "map[string]int32":
		v := m.data.(map[string]int32)
		return len(v)
	case "map[string]int64":
		v := m.data.(map[string]int64)
		return len(v)
	case "map[string]uint32":
		v := m.data.(map[string]uint32)
		return len(v)
	case "map[string]uint64":
		v := m.data.(map[string]uint64)
		return len(v)
	case "map[string]float32":
		v := m.data.(map[string]float32)
		return len(v)
	case "map[string]float64":
		v := m.data.(map[string]float64)
		return len(v)
	case "map[string]bool":
		v := m.data.(map[string]bool)
		return len(v)
	case "map[string]Meta":
		v := m.data.(map[string]Meta)
		return len(v)
	default:
		return 0
	}
}

func (m *Meta) String() string {
	switch m.data_type {
	case "string":
		return m.data.(string)
	case "int32", "int64", "uint32", "uint64", "float32", "float64":
		return fmt.Sprint(m.data)
	case "bool":
		if m.data.(bool) {
			return "true"
		}
		return "false"
	default:
		return ""
	}
}

func (m *Meta) Int32() int32 {
	switch m.data_type {
	case "string":
		v, err := strconv.ParseInt(m.data.(string), 10, 32)
		if err != nil {
			return 0
		}
		return int32(v)
	case "int32":
		return m.data.(int32)
	case "int64":
		return int32(m.data.(int64))
	case "uint32":
		return int32(m.data.(uint32))
	case "uint64":
		return int32(m.data.(uint64))
	case "float32":
		return int32(m.data.(float32))
	case "float64":
		return int32(m.data.(float64))
	case "bool":
		if m.data.(bool) {
			return 1
		}
		return 0
	default:
		return 0
	}
}

func (m *Meta) Int64() int64 {
	switch m.data_type {
	case "string":
		v, err := strconv.ParseInt(m.data.(string), 10, 64)
		if err != nil {
			return 0
		}
		return v
	case "int32":
		return int64(m.data.(int32))
	case "int64":
		return m.data.(int64)
	case "uint32":
		return int64(m.data.(uint32))
	case "uint64":
		return int64(m.data.(uint64))
	case "float32":
		return int64(m.data.(float32))
	case "float64":
		return int64(m.data.(float64))
	case "bool":
		if m.data.(bool) {
			return 1
		}
		return 0
	default:
		return 0
	}
}

func (m *Meta) Uint32() uint32 {
	switch m.data_type {
	case "string":
		v, err := strconv.ParseUint(m.data.(string), 10, 32)
		if err != nil {
			return 0
		}
		return uint32(v)
	case "int32":
		return uint32(m.data.(int32))
	case "int64":
		return uint32(m.data.(int64))
	case "uint32":
		return m.data.(uint32)
	case "uint64":
		return uint32(m.data.(uint64))
	case "float32":
		return uint32(m.data.(float32))
	case "float64":
		return uint32(m.data.(float64))
	case "bool":
		if m.data.(bool) {
			return 1
		}
		return 0
	default:
		return 0
	}
}

func (m *Meta) Uint64() uint64 {
	switch m.data_type {
	case "string":
		v, err := strconv.ParseUint(m.data.(string), 10, 64)
		if err != nil {
			return 0
		}
		return v
	case "int32":
		return uint64(m.data.(int32))
	case "int64":
		return uint64(m.data.(int64))
	case "uint32":
		return uint64(m.data.(uint32))
	case "uint64":
		return m.data.(uint64)
	case "float32":
		return uint64(m.data.(float32))
	case "float64":
		return uint64(m.data.(float64))
	case "bool":
		if m.data.(bool) {
			return 1
		}
		return 0
	default:
		return 0
	}
}

func (m *Meta) Float32() float32 {
	switch m.data_type {
	case "string":
		v, err := strconv.ParseFloat(m.data.(string), 32)
		if err != nil {
			return 0.0
		}
		return float32(v)
	case "int32":
		return float32(m.data.(int32))
	case "int64":
		return float32(m.data.(int64))
	case "uint32":
		return float32(m.data.(uint32))
	case "uint64":
		return float32(m.data.(uint64))
	case "float32":
		return m.data.(float32)
	case "float64":
		return float32(m.data.(float64))
	case "bool":
		if m.data.(bool) {
			return 1.0
		}
		return 0.0
	default:
		return 0.0
	}
}

func (m *Meta) Float64() float64 {
	switch m.data_type {
	case "string":
		v, err := strconv.ParseFloat(m.data.(string), 64)
		if err != nil {
			return 0.0
		}
		return v
	case "int32":
		return float64(m.data.(int32))
	case "int64":
		return float64(m.data.(int64))
	case "uint32":
		return float64(m.data.(uint32))
	case "uint64":
		return float64(m.data.(uint64))
	case "float32":
		return float64(m.data.(float32))
	case "float64":
		return m.data.(float64)
	case "bool":
		if m.data.(bool) {
			return 1.0
		}
		return 0.0
	default:
		return 0.0
	}
}

func (m *Meta) Bool() bool {
	switch m.data_type {
	case "string":
		v, err := strconv.ParseBool(m.data.(string))
		if err != nil {
			return false
		}
		return v
	case "int32":
		return m.data.(int32) != 0
	case "int64":
		return m.data.(int64) != 0
	case "uint32":
		return m.data.(uint32) != 0
	case "uint64":
		return m.data.(uint64) != 0
	case "float32":
		return m.data.(float32) != 0.0
	case "float64":
		return m.data.(float64) != 0.0
	case "bool":
		return m.data.(bool)
	default:
		return false
	}
}

func (m *Meta) VecString() []string {
	if m.data_type != "[]string" {
		return nil
	}
	return m.data.([]string)
}

func (m *Meta) VecInt32() []int32 {
	if m.data_type != "[]int32" {
		return nil
	}
	return m.data.([]int32)
}

func (m *Meta) VecInt64() []int64 {
	if m.data_type != "[]int64" {
		return nil
	}
	return m.data.([]int64)
}

func (m *Meta) VecUint32() []uint32 {
	if m.data_type != "[]uint32" {
		return nil
	}
	return m.data.([]uint32)
}

func (m *Meta) VecUint64() []uint64 {
	if m.data_type != "[]uint64" {
		return nil
	}
	return m.data.([]uint64)
}

func (m *Meta) VecFloat32() []float32 {
	if m.data_type != "[]float32" {
		return nil
	}
	return m.data.([]float32)
}

func (m *Meta) VecFloat64() []float64 {
	if m.data_type != "[]float64" {
		return nil
	}
	return m.data.([]float64)
}

func (m *Meta) VecBool() []bool {
	if m.data_type != "[]bool" {
		return nil
	}
	return m.data.([]bool)
}

func (m *Meta) VecMeta() []Meta {
	if m.data_type != "[]Meta" {
		return nil
	}
	return m.data.([]Meta)
}

func (m *Meta) MapString() map[string]string {
	if m.data_type != "map[string]string" {
		return nil
	}
	return m.data.(map[string]string)
}

func (m *Meta) MapInt32() map[string]int32 {
	if m.data_type != "map[string]int32" {
		return nil
	}
	return m.data.(map[string]int32)
}

func (m *Meta) MapInt64() map[string]int64 {
	if m.data_type != "map[string]int64" {
		return nil
	}
	return m.data.(map[string]int64)
}

func (m *Meta) MapUint32() map[string]uint32 {
	if m.data_type != "map[string]uint32" {
		return nil
	}
	return m.data.(map[string]uint32)
}

func (m *Meta) MapUint64() map[string]uint64 {
	if m.data_type != "map[string]uint64" {
		return nil
	}
	return m.data.(map[string]uint64)
}

func (m *Meta) MapFloat32() map[string]float32 {
	if m.data_type != "map[string]float32" {
		return nil
	}
	return m.data.(map[string]float32)
}

func (m *Meta) MapFloat64() map[string]float64 {
	if m.data_type != "map[string]float64" {
		return nil
	}
	return m.data.(map[string]float64)
}

func (m *Meta) MapBool() map[string]bool {
	if m.data_type != "map[string]bool" {
		return nil
	}
	return m.data.(map[string]bool)
}

func (m *Meta) MapMeta() map[string]Meta {
	if m.data_type != "map[string]Meta" {
		return nil
	}
	return m.data.(map[string]Meta)
}

func (m *Meta) MarshalJSON() ([]byte, error) {
	return json.Marshal(map[string]any{
		"data": m.data,
		"type": m.data_type,
	})
}

func NewMeta() *Meta {
	return &Meta{
		data:      nil,
		data_type: "",
	}
}

func NewMetaFrom(val any) *Meta {
	m := NewMeta()
	m.Set(val)
	return m
}

func NewMetaFromJSON(payload []byte) (*Meta, error) {
	in := map[string]any{}
	err := json.Unmarshal(payload, &in)
	if err != nil {
		return nil, err
	}
	data, has_data := in["data"]
	data_type, has_type := in["type"]
	if !has_data || !has_type {
		return nil, fmt.Errorf("missing 'data' and/or 'type'")
	}
	dtype, ok := data_type.(string)
	if !ok {
		return nil, fmt.Errorf("'type' must be string")
	}
	m := NewMeta()
	switch dtype {
	case "string", "int32", "int64", "uint32", "uint64", "float32", "float64", "bool":
		if !m.Set(data) {
			return nil, fmt.Errorf("'data' wasn't convertable")
		}
		return m, nil
	case "[]string":
		v := []string{}
		_, ok := data.([]any)[0].(string)
		if !ok {
			return nil, fmt.Errorf("'data' isn't []string")
		}
		for _, val := range data.([]any) {
			v = append(v, val.(string))
		}
		if !m.Set(v) {
			return nil, fmt.Errorf("'data' wasn't convertable")
		}
		return m, nil
	case "[]int32":
		v := []int32{}
		_, ok := data.([]any)[0].(int32)
		if !ok {
			return nil, fmt.Errorf("'data' isn't []int32")
		}
		for _, val := range data.([]any) {
			v = append(v, val.(int32))
		}
		if !m.Set(v) {
			return nil, fmt.Errorf("'data' wasn't convertable")
		}
		return m, nil
	case "[]int64":
		v := []int64{}
		_, ok := data.([]any)[0].(int64)
		if !ok {
			return nil, fmt.Errorf("'data' isn't []int64")
		}
		for _, val := range data.([]any) {
			v = append(v, val.(int64))
		}
		if !m.Set(v) {
			return nil, fmt.Errorf("'data' wasn't convertable")
		}
		return m, nil
	case "[]uint32":
		v := []uint32{}
		_, ok := data.([]any)[0].(uint32)
		if !ok {
			return nil, fmt.Errorf("'data' isn't []uint32")
		}
		for _, val := range data.([]any) {
			v = append(v, val.(uint32))
		}
		if !m.Set(v) {
			return nil, fmt.Errorf("'data' wasn't convertable")
		}
		return m, nil
	case "[]uint64":
		v := []uint64{}
		_, ok := data.([]any)[0].(uint64)
		if !ok {
			return nil, fmt.Errorf("'data' isn't []uint64")
		}
		for _, val := range data.([]any) {
			v = append(v, val.(uint64))
		}
		if !m.Set(v) {
			return nil, fmt.Errorf("'data' wasn't convertable")
		}
		return m, nil
	case "[]float32":
		v := []float32{}
		_, ok := data.([]any)[0].(float32)
		if !ok {
			return nil, fmt.Errorf("'data' isn't []float32")
		}
		for _, val := range data.([]any) {
			v = append(v, val.(float32))
		}
		if !m.Set(v) {
			return nil, fmt.Errorf("'data' wasn't convertable")
		}
		return m, nil
	case "[]float64":
		v := []float64{}
		_, ok := data.([]any)[0].(float64)
		if !ok {
			return nil, fmt.Errorf("'data' isn't []float64")
		}
		for _, val := range data.([]any) {
			v = append(v, val.(float64))
		}
		if !m.Set(v) {
			return nil, fmt.Errorf("'data' wasn't convertable")
		}
		return m, nil
	case "[]bool":
		v := []bool{}
		_, ok := data.([]any)[0].(bool)
		if !ok {
			return nil, fmt.Errorf("'data' isn't []bool")
		}
		for _, val := range data.([]any) {
			v = append(v, val.(bool))
		}
		if !m.Set(v) {
			return nil, fmt.Errorf("'data' wasn't convertable")
		}
		return m, nil
	case "[]Meta":
		v := []Meta{}
		for _, val := range data.([]any) {
			i, err := json.Marshal(val.(map[string]any))
			if err != nil {
				return nil, err
			}
			om, err := NewMetaFromJSON(i)
			if err != nil {
				return nil, err
			}
			v = append(v, *om)
		}
		if !m.Set(v) {
			return nil, fmt.Errorf("'data' wasn't convertable")
		}
		return m, nil
	case "map[string]string":
		v := map[string]string{}
		for key, val := range data.(map[string]any) {
			v[key] = val.(string)
		}
		if !m.Set(v) {
			return nil, fmt.Errorf("'data' wasn't convertable")
		}
		return m, nil
	case "map[string]int32":
		v := map[string]int32{}
		for key, val := range data.(map[string]any) {
			v[key] = val.(int32)
		}
		if !m.Set(v) {
			return nil, fmt.Errorf("'data' wasn't convertable")
		}
		return m, nil
	case "map[string]int64":
		v := map[string]int64{}
		for key, val := range data.(map[string]any) {
			v[key] = val.(int64)
		}
		if !m.Set(v) {
			return nil, fmt.Errorf("'data' wasn't convertable")
		}
		return m, nil
	case "map[string]uint32":
		v := map[string]uint32{}
		for key, val := range data.(map[string]any) {
			v[key] = val.(uint32)
		}
		if !m.Set(v) {
			return nil, fmt.Errorf("'data' wasn't convertable")
		}
		return m, nil
	case "map[string]uint64":
		v := map[string]uint64{}
		for key, val := range data.(map[string]any) {
			v[key] = val.(uint64)
		}
		if !m.Set(v) {
			return nil, fmt.Errorf("'data' wasn't convertable")
		}
		return m, nil
	case "map[string]float32":
		v := map[string]float32{}
		for key, val := range data.(map[string]any) {
			v[key] = val.(float32)
		}
		if !m.Set(v) {
			return nil, fmt.Errorf("'data' wasn't convertable")
		}
		return m, nil
	case "map[string]float64":
		v := map[string]float64{}
		for key, val := range data.(map[string]any) {
			v[key] = val.(float64)
		}
		if !m.Set(v) {
			return nil, fmt.Errorf("'data' wasn't convertable")
		}
		return m, nil
	case "map[string]bool":
		v := map[string]bool{}
		for key, val := range data.(map[string]any) {
			v[key] = val.(bool)
		}
		if !m.Set(v) {
			return nil, fmt.Errorf("'data' wasn't convertable")
		}
		return m, nil
	case "map[string]Meta":
		v := map[string]Meta{}
		for key, val := range data.(map[string]any) {
			i, err := json.Marshal(val.(map[string]any))
			if err != nil {
				return nil, err
			}
			om, err := NewMetaFromJSON(i)
			if err != nil {
				return nil, err
			}
			v[key] = *om
		}
		if !m.Set(v) {
			return nil, fmt.Errorf("'data' wasn't convertable")
		}
		return m, nil
	default:
		return nil, fmt.Errorf("'data' wasn't convertable")
	}
}
