package main

import "strings"

type MetaData struct {
	metadata map[string]*Meta
	safe     bool
}

func NewMetaData() *MetaData {
	return &MetaData{
		metadata: make(map[string]*Meta),
		safe:     true,
	}
}

func (m *MetaData) Safe() bool {
	return m.safe
}

func (m *MetaData) Unsafe() {
	m.safe = false
}

func (m *MetaData) Set(key string, val any) {
	if !m.Has(key) {
		m.metadata[key] = NewMetaFrom(val)
		return
	}
	if m.Private(key) || m.Restricted(key) {
		return
	}
	m.metadata[key].Set(val)
}

func (m *MetaData) Has(key string) bool {
	return m.metadata[key] != nil
}

func (m *MetaData) Private(key string) bool {
	return strings.HasPrefix(key, ".") && m.Has(key) && !m.safe
}

func (m *MetaData) Restricted(key string) bool {
	return strings.HasPrefix(key, "*") && m.Has(key) && !m.safe
}

func (m *MetaData) Type(key string) string {
	if !m.Has(key) {
		return ""
	}
	return m.metadata[key].data_type
}

func (m *MetaData) String(key string) string {
	if m.Restricted(key) {
		return ""
	}
	return m.metadata[key].String()
}

func (m *MetaData) Int32(key string) int32 {
	if m.Restricted(key) {
		return 0
	}
	return m.metadata[key].Int32()
}

func (m *MetaData) Int64(key string) int64 {
	if m.Restricted(key) {
		return 0
	}
	return m.metadata[key].Int64()
}

func (m *MetaData) Uint32(key string) uint32 {
	if m.Restricted(key) {
		return 0
	}
	return m.metadata[key].Uint32()
}

func (m *MetaData) Uint64(key string) uint64 {
	if m.Restricted(key) {
		return 0
	}
	return m.metadata[key].Uint64()
}

func (m *MetaData) Float32(key string) float32 {
	if m.Restricted(key) {
		return 0.0
	}
	return m.metadata[key].Float32()
}

func (m *MetaData) Float64(key string) float64 {
	if m.Restricted(key) {
		return 0.0
	}
	return m.metadata[key].Float64()
}

func (m *MetaData) Bool(key string) bool {
	if m.Restricted(key) {
		return false
	}
	return m.metadata[key].Bool()
}

func (m *MetaData) VecString(key string) []string {
	if m.Restricted(key) {
		return nil
	}
	return m.metadata[key].VecString()
}

func (m *MetaData) VecInt32(key string) []int32 {
	if m.Restricted(key) {
		return nil
	}
	return m.metadata[key].VecInt32()
}

func (m *MetaData) VecInt64(key string) []int64 {
	if m.Restricted(key) {
		return nil
	}
	return m.metadata[key].VecInt64()
}

func (m *MetaData) VecUint32(key string) []uint32 {
	if m.Restricted(key) {
		return nil
	}
	return m.metadata[key].VecUint32()
}

func (m *MetaData) VecUint64(key string) []uint64 {
	if m.Restricted(key) {
		return nil
	}
	return m.metadata[key].VecUint64()
}

func (m *MetaData) VecFloat32(key string) []float32 {
	if m.Restricted(key) {
		return nil
	}
	return m.metadata[key].VecFloat32()
}

func (m *MetaData) VecFloat64(key string) []float64 {
	if m.Restricted(key) {
		return nil
	}
	return m.metadata[key].VecFloat64()
}

func (m *MetaData) VecBool(key string) []bool {
	if m.Restricted(key) {
		return nil
	}
	return m.metadata[key].VecBool()
}

func (m *MetaData) VecMeta(key string) []Meta {
	if m.Restricted(key) {
		return nil
	}
	return m.metadata[key].VecMeta()
}

func (m *MetaData) MapString(key string) map[string]string {
	if m.Restricted(key) {
		return nil
	}
	return m.metadata[key].MapString()
}

func (m *MetaData) MapInt32(key string) map[string]int32 {
	if m.Restricted(key) {
		return nil
	}
	return m.metadata[key].MapInt32()
}

func (m *MetaData) MapInt64(key string) map[string]int64 {
	if m.Restricted(key) {
		return nil
	}
	return m.metadata[key].MapInt64()
}

func (m *MetaData) MapUint32(key string) map[string]uint32 {
	if m.Restricted(key) {
		return nil
	}
	return m.metadata[key].MapUint32()
}

func (m *MetaData) MapUint64(key string) map[string]uint64 {
	if m.Restricted(key) {
		return nil
	}
	return m.metadata[key].MapUint64()
}

func (m *MetaData) MapFloat32(key string) map[string]float32 {
	if m.Restricted(key) {
		return nil
	}
	return m.metadata[key].MapFloat32()
}

func (m *MetaData) MapFloat64(key string) map[string]float64 {
	if m.Restricted(key) {
		return nil
	}
	return m.metadata[key].MapFloat64()
}

func (m *MetaData) MapBool(key string) map[string]bool {
	if m.Restricted(key) {
		return nil
	}
	return m.metadata[key].MapBool()
}

func (m *MetaData) MapMeta(key string) map[string]Meta {
	if m.Restricted(key) {
		return nil
	}
	return m.metadata[key].MapMeta()
}
